{% extends "admin/change_form.html" %}
{% block extrastyle %}{{ block.super }}
<style> 
    .readonly {
        overflow-y: scroll;
    }
</style>
{% endblock %}
{% block object-tools-items %}{% endblock %}